@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2021 Artifex Software, Inc.
+// Copyright (C) 2004-2025 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -29,5 +29,13 @@ public interface StructuredTextWalker
 	void endTextBlock();
 	void beginLine(Rect bbox, int wmode, Point dir);
 	void endLine();
-	void onChar(int c, Point origin, Font font, float size, Quad q);
+	void onChar(int c, Point origin, Font font, float size, Quad q, int argb, int flags);
+	void beginStruct(String standard, String raw, int index);
+	void endStruct();
+	void onVector(Rect bbox, VectorInfo info, int argb);
+
+	public static class VectorInfo {
+		public boolean isStroked;
+		public boolean isRectangle;
+	}
 }
