@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2025 Artifex Software, Inc.
+// Copyright (C) 2004-2026 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -122,10 +122,15 @@ typedef struct
 	fz_stext_page *page;
 	int id;
 	fz_point pen, start;
+	// maybe_bullet: True if the 'start' position recorded was done so after either some actualtext
+	// on an image, or after a glyph that's known to be used for bullets. This is used to stop us
+	// spotting an 'indented' paragraph, because it's possibly just a bulleted list.
+	int maybe_bullet;
 	fz_point lag_pen;
 	fz_matrix trm;
 	int new_obj;
 	int lastchar;
+	fz_stext_line *lastline;
 	int lastbidi;
 	int flags;
 	int color;
@@ -151,10 +156,13 @@ typedef struct
 	int rect_max;
 	int rect_len;
 	rect_details *rects;
+
+	fz_stext_block *lazy_vectors;
+	fz_stext_block *lazy_vectors_tail;
 } fz_stext_device;
 
 const char *fz_stext_options_usage =
-	"Text output options:\n"
+	"Structured text options:\n"
 	"\tpreserve-images: keep images in output\n"
 	"\tpreserve-ligatures: do not expand ligatures into constituent characters\n"
 	"\tpreserve-spans: do not merge spans on the same line\n"
@@ -175,6 +183,7 @@ const char *fz_stext_options_usage =
 	"\tvectors: include vector bboxes in output\n"
 	"\tsegment: attempt to segment the page\n"
 	"\ttable-hunt: hunt for tables within a (segmented) page\n"
+	"\tresolution: resolution to render at\n"
 	"\n";
 
 /* Find the current actualtext, if any. Will abort if dev == NULL. */
@@ -321,9 +330,8 @@ add_block_to_page(fz_context *ctx, fz_stext_page *page, int type, int id)
 	}
 	else if (!page->last_block)
 	{
-		page->last_block = block;
-		if (!page->first_block)
-			page->first_block = block;
+		assert(!page->first_block);
+		page->first_block = page->last_block = block;
 	}
 	else
 	{
@@ -333,6 +341,59 @@ add_block_to_page(fz_context *ctx, fz_stext_page *page, int type, int id)
 	return block;
 }
 
+static fz_stext_block *
+add_lazy_vector(fz_context *ctx, fz_stext_page *page, fz_stext_device *tdev, int id)
+{
+	fz_stext_block *block = fz_pool_alloc(ctx, page->pool, sizeof *page->first_block);
+	block->bbox = fz_empty_rect;
+	block->prev = tdev->lazy_vectors_tail;
+	block->type = FZ_STEXT_BLOCK_VECTOR;
+	block->id = id;
+
+	if (tdev->lazy_vectors == NULL)
+		tdev->lazy_vectors = block;
+	else
+		tdev->lazy_vectors_tail->next = block;
+	tdev->lazy_vectors_tail = block;
+
+	return block;
+}
+
+static void
+flush_lazy_vectors(fz_context *ctx, fz_stext_page *page, fz_stext_device *tdev)
+{
+	if (tdev->lazy_vectors == NULL)
+		return;
+
+	if (page->last_struct)
+	{
+		if (page->last_struct->last_block)
+		{
+			page->last_struct->last_block->next = tdev->lazy_vectors;
+			tdev->lazy_vectors->prev = page->last_struct->last_block;
+			page->last_struct->last_block = tdev->lazy_vectors_tail;
+		}
+		else
+		{
+			page->last_struct->first_block = tdev->lazy_vectors;
+			page->last_struct->last_block = tdev->lazy_vectors_tail;
+		}
+	}
+	else if (!page->last_block)
+	{
+		page->first_block = tdev->lazy_vectors;
+		page->last_block = tdev->lazy_vectors_tail;
+	}
+	else
+	{
+		page->last_block->next = tdev->lazy_vectors;
+		tdev->lazy_vectors->prev = page->last_block;
+		page->last_block = tdev->lazy_vectors_tail;
+	}
+
+	tdev->lazy_vectors = tdev->lazy_vectors_tail = NULL;
+}
+
 static fz_stext_block *
 add_text_block_to_page(fz_context *ctx, fz_stext_page *page, int id)
 {
@@ -391,7 +452,7 @@ add_char_to_line(fz_context *ctx, fz_stext_page *page, fz_stext_line *line, fz_m
 	ch->origin = *p;
 	ch->size = size;
 	ch->font = fz_keep_font(ctx, font);
-	ch->flags = flags | (synthetic ? FZ_STEXT_SYNTHETIC : 0);
+	ch->flags = flags | (synthetic ? FZ_STEXT_SYNTHETIC : 0) | (synthetic > 1 ? FZ_STEXT_SYNTHETIC_LARGE : 0);
 	if (font->flags.is_bold)
 		ch->flags |= FZ_STEXT_BOLD;
 
@@ -448,29 +509,6 @@ add_char_to_line(fz_context *ctx, fz_stext_page *page, fz_stext_line *line, fz_m
 	return ch;
 }
 
-static void
-remove_last_char(fz_context *ctx, fz_stext_line *line)
-{
-	if (line && line->first_char)
-	{
-		fz_stext_char *prev = NULL;
-		fz_stext_char *ch = line->first_char;
-		while (ch->next)
-		{
-			prev = ch;
-			ch = ch->next;
-		}
-		if (prev)
-		{
-			/* The characters are pool allocated, so we don't actually leak the removed node. */
-			/* We do need to drop the char's font reference though. */
-			fz_drop_font(ctx, prev->next->font);
-			line->last_char = prev;
-			line->last_char->next = NULL;
-		}
-	}
-}
-
 static fz_stext_char *reverse_bidi_span(fz_stext_char *curr, fz_stext_char *tail)
 {
 	fz_stext_char *prev, *next;
@@ -504,7 +542,7 @@ static void reverse_bidi_line(fz_stext_line *line)
 	}
 }
 
-static int is_hyphen(int c)
+int fz_is_unicode_hyphen(int c)
 {
 	/* check for: hyphen-minus, soft hyphen, hyphen, and non-breaking hyphen */
 	return (c == '-' || c == 0xAD || c == 0x2010 || c == 0x2011);
@@ -526,10 +564,10 @@ static int may_add_space(int lastchar)
 	return (lastchar != ' ' && (lastchar < 0x700 || (lastchar >= 0x2000 && lastchar <= 0x20CF)));
 }
 
-#define FAKEBOLD_THRESHOLD_RECIP 10
+#define FAKEBOLD_THRESHOLD_RECIP (1.0f / FAKE_BOLD_MAX_DIST)
 
 static int
-close(float a, float b, float size)
+is_within_fake_bold_distance(float a, float b, float size)
 {
 	a -= b;
 	if (a < 0)
@@ -550,6 +588,9 @@ font_equiv(fz_context *ctx, fz_font *f, fz_font *g)
 	if (strcmp(f->name, g->name) != 0)
 		return 0;
 
+	if (f->buffer == NULL || g->buffer == NULL)
+		return 0;
+
 	fz_font_digest(ctx, f, fdigest);
 	fz_font_digest(ctx, g, gdigest);
 
@@ -577,7 +618,7 @@ check_for_fake_bold(fz_context *ctx, fz_stext_block *block, fz_font *font, int c
 				for (ch = line->first_char; ch != NULL; ch = ch->next)
 				{
 					/* Not perfect, but it'll do! */
-					if (ch->c == c && close(ch->origin.x, p.x, size) && close(ch->origin.y, p.y, size) && font_equiv(ctx, ch->font, font))
+					if (ch->c == c && is_within_fake_bold_distance(ch->origin.x, p.x, size) && is_within_fake_bold_distance(ch->origin.y, p.y, size) && font_equiv(ctx, ch->font, font))
 					{
 						/* If we were filled before, and we are stroking now... */
 						if ((ch->flags & (FZ_STEXT_FILLED | FZ_STEXT_STROKED)) == FZ_STEXT_FILLED &&
@@ -598,9 +639,10 @@ check_for_fake_bold(fz_context *ctx, fz_stext_block *block, fz_font *font, int c
 							{
 								/* OK, we can be bold. */
 								ch->flags |= FZ_STEXT_BOLD;
-								return 1;
 							}
-							/* Ignore this and keep going */
+							/* Whether we have recorded this as being bold or not, still
+							 * claim we did, so we swallow the space and don't reemit it. */
+							return 1;
 						}
 						else
 						{
@@ -617,12 +659,71 @@ check_for_fake_bold(fz_context *ctx, fz_stext_block *block, fz_font *font, int c
 	return 0;
 }
 
+static int
+plausible_bullet(int c)
+{
+	return (c == '*' ||
+		c == 0x00B7 || /* Middle Dot */
+		c == 0x2022 || /* Bullet */
+		c == 0x2023 || /* Triangular Bullet */
+		c == 0x2043 || /* Hyphen Bullet */
+		c == 0x204C || /* Back leftwards bullet */
+		c == 0x204D || /* Back rightwards bullet */
+		c == 0x2219 || /* Bullet operator */
+		c == 0x25C9 || /* Fisheye */
+		c == 0x25CB || /* White circle */
+		c == 0x25CF || /* Black circle */
+		c == 0x25D8 || /* Inverse Bullet */
+		c == 0x25E6 || /* White Bullet */
+		c == 0x2619 || /* Reversed Rotated Floral Heart Bullet / Fleuron */
+		c == 0x261a || /* Black left pointing index */
+		c == 0x261b || /* Black right pointing index */
+		c == 0x261c || /* White left pointing index */
+		c == 0x261d || /* White up pointing index */
+		c == 0x261e || /* White right pointing index */
+		c == 0x261f || /* White down pointing index */
+		c == 0x2765 || /* Rotated Heavy Heart Black Heart Bullet */
+		c == 0x2767 || /* Rotated Floral Heart Bullet / Fleuron */
+		c == 0x29BE || /* Circled White Bullet */
+		c == 0x29BF || /* Circled Bullet */
+		c == 0x2660 || /* Black Spade suit */
+		c == 0x2661 || /* White Heart suit */
+		c == 0x2662 || /* White Diamond suit */
+		c == 0x2663 || /* Black Club suit */
+		c == 0x2664 || /* White Spade suit */
+		c == 0x2665 || /* Black Heart suit */
+		c == 0x2666 || /* Black Diamond suit */
+		c == 0x2667 || /* White Clud suit */
+		c == 0x1F446 || /* WHITE UP POINTING BACKHAND INDEX */
+		c == 0x1F447 || /* WHITE DOWN POINTING BACKHAND INDEX */
+		c == 0x1F448 || /* WHITE LEFT POINTING BACKHAND INDEX */
+		c == 0x1F449 || /* WHITE RIGHT POINTING BACKHAND INDEX */
+		c == 0x1f597 || /* White down pointing left hand index */
+		c == 0x1F598 || /* SIDEWAYS WHITE LEFT POINTING INDEX */
+		c == 0x1F599 || /* SIDEWAYS WHITE RIGHT POINTING INDEX */
+		c == 0x1F59A || /* SIDEWAYS BLACK LEFT POINTING INDEX */
+		c == 0x1F59B || /* SIDEWAYS BLACK RIGHT POINTING INDEX */
+		c == 0x1F59C || /* BLACK LEFT POINTING BACKHAND INDEX */
+		c == 0x1F59D || /* BLACK RIGHT POINTING BACKHAND INDEX */
+		c == 0x1F59E || /* SIDEWAYS WHITE UP POINTING INDEX */
+		c == 0x1F59F || /* SIDEWAYS WHITE DOWN POINTING INDEX */
+		c == 0x1F5A0 || /* SIDEWAYS BLACK UP POINTING INDEX */
+		c == 0x1F5A1 || /* SIDEWAYS BLACK DOWN POINTING INDEX */
+		c == 0x1F5A2 || /* BLACK UP POINTING BACKHAND INDEX */
+		c == 0x1F5A3 || /* BLACK DOWN POINTING BACKHAND INDEX */
+		c == 0x1FBC1 || /* LEFT THIRD WHITE RIGHT POINTING INDEX */
+		c == 0x1FBC2 || /* MIDDLE THIRD WHITE RIGHT POINTING INDEX */
+		c == 0x1FBC3 || /* RIGHT THIRD WHITE RIGHT POINTING INDEX */
+		c == 0xFFFD || /* UNICODE_REPLACEMENT_CHARACTER */
+		0);
+}
+
 static void
 fz_add_stext_char_imp(fz_context *ctx, fz_stext_device *dev, fz_font *font, int c, int glyph, fz_matrix trm, float adv, int wmode, int bidi, int force_new_line, int flags)
 {
 	fz_stext_page *page = dev->page;
 	fz_stext_block *cur_block;
-	fz_stext_line *cur_line;
+	fz_stext_line *cur_line = NULL;
 
 	int new_para = 0;
 	int new_line = 1;
@@ -683,17 +784,19 @@ fz_add_stext_char_imp(fz_context *ctx, fz_stext_device *dev, fz_font *font, int
 		q.y = trm.f;
 	}
 
+	//printf("%g,%g \"%c\" %g,%g\n", p.x, p.y, c, q.x, q.y);
+
 	if ((dev->opts.flags & FZ_STEXT_COLLECT_STYLES) != 0)
 	{
 		if (glyph == -1)
 		{
 			if (dev->last_was_fake_bold)
-				goto move_pen_and_exit;
+				return;
 		}
 		else if (check_for_fake_bold(ctx, page->first_block, font, c, p, size, flags))
 		{
 			dev->last_was_fake_bold = 1;
-			goto move_pen_and_exit;
+			return;
 		}
 		dev->last_was_fake_bold = 0;
 	}
@@ -704,12 +807,15 @@ fz_add_stext_char_imp(fz_context *ctx, fz_stext_device *dev, fz_font *font, int
 		cur_block = NULL;
 	cur_line = cur_block ? cur_block->u.t.last_line : NULL;
 
-	if (cur_line && glyph < 0)
+	/* We use glyph == -2 to indicate a no-glyph char from an actualtext. The position
+	 * is valid though, so we want to advance the pen for these. */
+	if (cur_line && glyph == -1)
 	{
 		/* Don't advance pen or break lines for no-glyph characters in a cluster */
 		add_char_to_line(ctx, page, cur_line, trm, font, size, c, (dev->flags & FZ_STEXT_ACCURATE_BBOXES) ? glyph : NON_ACCURATE_GLYPH, &dev->pen, &dev->pen, bidi, dev->color, 0, flags, dev->flags);
 		dev->lastbidi = bidi;
 		dev->lastchar = c;
+		dev->lastline = cur_line;
 		return;
 	}
 
@@ -726,8 +832,10 @@ fz_add_stext_char_imp(fz_context *ctx, fz_stext_device *dev, fz_font *font, int
 		/* Largely supplanted by the check_for_fake_bold mechanism above,
 		 * but we leave this in for backward compatibility as it's cheap,
 		 * and works even when FZ_STEXT_COLLECT_STYLES is not set. */
-		dist = hypotf(q.x - dev->pen.x, q.y - dev->pen.y) / size;
-		if (dist < FAKE_BOLD_MAX_DIST && c == dev->lastchar)
+		dist = hypotf(p.x - dev->lag_pen.x, p.y - dev->lag_pen.y) / size;
+		/* This can trigger improperly for glyphs that come from actualtext
+		 * as they are frequently overlaid. Therefore rely on glyph >= 0. */
+		if (dist < FAKE_BOLD_MAX_DIST && c == dev->lastchar && glyph >= 0)
 			return;
 
 		/* Calculate how far we've moved since the last character. */
@@ -795,7 +903,7 @@ fz_add_stext_char_imp(fz_context *ctx, fz_stext_device *dev, fz_font *font, int
 				{
 					bidi = 3; /* mark line as visual */
 					if (wmode == 0 && may_add_space(dev->lastchar))
-						add_space = 1;
+						add_space = 1 + (spacing > SPACE_DIST*2);
 					new_line = 0;
 				}
 
@@ -824,7 +932,7 @@ fz_add_stext_char_imp(fz_context *ctx, fz_stext_device *dev, fz_font *font, int
 				{
 					/* Motion is forward in line and large enough to warrant us adding a space. */
 					if (wmode == 0 && may_add_space(dev->lastchar))
-						add_space = 1;
+						add_space = 1 + (spacing > SPACE_DIST*2);
 					new_line = 0;
 				}
 				else
@@ -840,7 +948,7 @@ fz_add_stext_char_imp(fz_context *ctx, fz_stext_device *dev, fz_font *font, int
 		{
 			/* Check indent to spot text-indent style paragraphs */
 			if (wmode == 0 && cur_line && dev->new_obj)
-				if ((p.x - dev->start.x) > 0.5f)
+				if ((p.x - dev->start.x) > 0.5f && !dev->maybe_bullet)
 					new_para = 1;
 			new_line = 1;
 		}
@@ -856,32 +964,38 @@ fz_add_stext_char_imp(fz_context *ctx, fz_stext_device *dev, fz_font *font, int
 	/* Start a new block (but only at the beginning of a text object) */
 	if (new_para || !cur_block)
 	{
+		flush_lazy_vectors(ctx, page, dev);
 		cur_block = add_text_block_to_page(ctx, page, dev->id);
 		cur_line = cur_block->u.t.last_line;
 	}
 
-	if (new_line && (dev->flags & FZ_STEXT_DEHYPHENATE) && is_hyphen(dev->lastchar))
-	{
-		remove_last_char(ctx, cur_line);
-		new_line = 0;
-	}
+	if (new_line && (dev->flags & FZ_STEXT_DEHYPHENATE) && fz_is_unicode_hyphen(dev->lastchar) && dev->lastline != NULL)
+		dev->lastline->flags |= FZ_STEXT_LINE_FLAGS_JOINED;
 
 	/* Start a new line */
 	if (new_line || !cur_line || force_new_line)
 	{
 		cur_line = add_line_to_block(ctx, page, cur_block, &ndir, wmode, bidi);
 		dev->start = p;
+		if (glyph == -2)
+			dev->maybe_bullet = 1;
+		else
+			dev->maybe_bullet = plausible_bullet(c);
 	}
 
+	/* Henceforth treat such non-glyphs in the usual way. */
+	if (glyph == -2)
+		glyph = -1;
+
 	/* Add synthetic space */
 	if (add_space && !(dev->flags & FZ_STEXT_INHIBIT_SPACES))
-		add_char_to_line(ctx, page, cur_line, trm, font, size, ' ', (dev->flags & FZ_STEXT_ACCURATE_BBOXES) ? NON_ACCURATE_GLYPH_ADDED_SPACE : NON_ACCURATE_GLYPH, &dev->pen, &p, bidi, dev->color, 1, flags, dev->flags);
+		add_char_to_line(ctx, page, cur_line, trm, font, size, ' ', (dev->flags & FZ_STEXT_ACCURATE_BBOXES) ? NON_ACCURATE_GLYPH_ADDED_SPACE : NON_ACCURATE_GLYPH, &dev->pen, &p, bidi, dev->color, add_space, flags, dev->flags);
 
 	add_char_to_line(ctx, page, cur_line, trm, font, size, c, (dev->flags & FZ_STEXT_ACCURATE_BBOXES) ? glyph : NON_ACCURATE_GLYPH, &p, &q, bidi, dev->color, 0, flags, dev->flags);
 
-move_pen_and_exit:
 	dev->lastchar = c;
 	dev->lastbidi = bidi;
+	dev->lastline = cur_line;
 	dev->lag_pen = p;
 	dev->pen = q;
 
@@ -1077,15 +1191,37 @@ rune_index(const char *utf8, size_t idx)
 }
 
 static void
-flush_actualtext(fz_context *ctx, fz_stext_device *dev, const char *actualtext, int i, int end)
+flush_actualtext(fz_context *ctx, fz_stext_device *dev, const char *actualtext, int i, int end, float adv)
 {
 	if (*actualtext == 0)
 		return;
 
+	if (!dev->last.valid)
+		return;
+
 	if (dev->flags & (FZ_STEXT_CLIP | FZ_STEXT_CLIP_RECT))
 		if (dev->last.clipped)
 			return;
 
+	if (adv != 0)
+	{
+		const char *at = actualtext;
+		int j = i;
+
+		while (end < 0 || (end >= 0 && i < end))
+		{
+			int rune;
+			at += fz_chartorune(&rune, at);
+
+			if (rune == 0)
+				break;
+			j++;
+		}
+
+		if (j != i)
+			adv /= (j - i);
+	}
+
 	while (end < 0 || (end >= 0 && i < end))
 	{
 		int rune;
@@ -1094,11 +1230,14 @@ flush_actualtext(fz_context *ctx, fz_stext_device *dev, const char *actualtext,
 		if (rune == 0)
 			break;
 
+		dev->last.trm.e = dev->pen.x;
+		dev->last.trm.f = dev->pen.y;
+
 		fz_add_stext_char(ctx, dev, dev->last.font,
 			rune,
-			-1,
+			-2,
 			dev->last.trm,
-			0,
+			adv,
 			dev->last.wmode,
 			dev->last.bidi_level,
 			(i == 0) && (dev->flags & FZ_STEXT_PRESERVE_SPANS),
@@ -1234,9 +1373,16 @@ do_extract_within_actualtext(fz_context *ctx, fz_stext_device *dev, fz_text_span
 		return;
 	}
 
+	/* if this is the first text on the page, and the actual text suffix matches the entire
+	 * span text, then no font will have been set above, so set the last used font to the
+	 * span font since flush_actualtext() assumes that a font has been set.
+	 */
+	if (!dev->last.font)
+		dev->last.font = fz_keep_font(ctx, font);
+
 	/* We found a matching postfix. It seems likely that this is going to be the only
 	 * text object we get, so send any remaining actualtext now. */
-	flush_actualtext(ctx, dev, actualtext, i, i + (int)strlen(actualtext) - (span->len - end));
+	flush_actualtext(ctx, dev, actualtext, i, i + (int)strlen(actualtext) - (span->len - end), 0);
 
 	/* Send the postfix */
 	if (end != span->len)
@@ -1269,10 +1415,10 @@ static uint32_t hexrgba_from_color(fz_context *ctx, fz_colorspace *colorspace, c
 	float rgb[3];
 	fz_convert_color(ctx, colorspace, color, fz_device_rgb(ctx), rgb, NULL, fz_default_color_params);
 	return
-		((uint32_t) (fz_clampi(alpha * 255 + 0.5f, 0, 255) << 24)) |
-		((uint32_t) (fz_clampi(rgb[0] * 255 + 0.5f, 0, 255) << 16)) |
-		((uint32_t) (fz_clampi(rgb[1] * 255 + 0.5f, 0, 255) << 8)) |
-		((uint32_t) (fz_clampi(rgb[2] * 255 + 0.5f, 0, 255)));
+		(((uint32_t) fz_clampi(alpha * 255 + 0.5f, 0, 255)) << 24) |
+		(((uint32_t) fz_clampi(rgb[0] * 255 + 0.5f, 0, 255)) << 16) |
+		(((uint32_t) fz_clampi(rgb[1] * 255 + 0.5f, 0, 255)) << 8) |
+		(((uint32_t) fz_clampi(rgb[2] * 255 + 0.5f, 0, 255)));
 }
 
 static void
@@ -1357,20 +1503,31 @@ fz_stext_begin_metatext(fz_context *ctx, fz_device *dev, fz_metatext meta, const
 {
 	fz_stext_device *tdev = (fz_stext_device*)dev;
 	metatext_t *mt = find_actualtext(tdev);
+	char *new_text = NULL;
 
 	if (mt != NULL && meta == FZ_METATEXT_ACTUALTEXT)
-		flush_actualtext(ctx, tdev, mt->text, 0, -1);
+		flush_actualtext(ctx, tdev, mt->text, 0, -1, 0);
 
 	if (meta == FZ_METATEXT_ACTUALTEXT)
 		tdev->last.valid = 0;
 
-	mt = fz_malloc_struct(ctx, metatext_t);
+	new_text = text ? fz_strdup(ctx, text) : NULL;
 
-	mt->prev = tdev->metatext;
-	tdev->metatext = mt;
-	mt->type = meta;
-	mt->text = text ? fz_strdup(ctx, text) : NULL;
-	mt->bounds = fz_empty_rect;
+	fz_try(ctx)
+	{
+		mt = fz_malloc_struct(ctx, metatext_t);
+
+		mt->prev = tdev->metatext;
+		tdev->metatext = mt;
+		mt->type = meta;
+		mt->text = new_text;
+		mt->bounds = fz_empty_rect;
+	}
+	fz_catch(ctx)
+	{
+		fz_free(ctx, new_text);
+		fz_rethrow(ctx);
+	}
 }
 
 static void
@@ -1400,7 +1557,7 @@ fz_stext_end_metatext(fz_context *ctx, fz_device *dev)
 	if (!tdev->metatext)
 		return; /* Mismatched pop. Live with it. */
 
-	if (tdev->metatext->type != FZ_METATEXT_ACTUALTEXT)
+	if (tdev->metatext->type != FZ_METATEXT_ACTUALTEXT || (tdev->opts.flags & FZ_STEXT_IGNORE_ACTUALTEXT) != 0)
 	{
 		/* We only deal with ActualText here. Just pop anything else off,
 		 * and we're done. */
@@ -1411,31 +1568,31 @@ fz_stext_end_metatext(fz_context *ctx, fz_device *dev)
 	/* If we have a 'last' text position, send the content after that. */
 	if (tdev->last.valid)
 	{
-		flush_actualtext(ctx, tdev, tdev->metatext->text, 0, -1);
+		flush_actualtext(ctx, tdev, tdev->metatext->text, 0, -1, 0);
 		pop_metatext(ctx, tdev);
 		tdev->last.valid = 0;
 		return;
 	}
 
-	/* If we have collected a rectangle for content that encloses the actual text,
-	 * send the content there. */
-	if (!fz_is_empty_rect(tdev->metatext->bounds))
+	/* Unless we have collected a rectangle for content that encloses the actual text,
+	 * we can't do anything. */
+	if (fz_is_empty_rect(tdev->metatext->bounds))
 	{
-		tdev->last.trm.a = tdev->metatext->bounds.x1 - tdev->metatext->bounds.x0;
-		tdev->last.trm.b = 0;
-		tdev->last.trm.c = 0;
-		tdev->last.trm.d = tdev->metatext->bounds.y1 - tdev->metatext->bounds.y0;
-		tdev->last.trm.e = tdev->metatext->bounds.x0;
-		tdev->last.trm.f = tdev->metatext->bounds.y0;
-	}
-	else
-	{
-		if ((dev->flags & (FZ_STEXT_CLIP | FZ_STEXT_CLIP_RECT)) == 0)
+		if ((dev->flags & (FZ_STEXT_CLIP | FZ_STEXT_CLIP_RECT)) == 0 && tdev->metatext->text[0])
 			fz_warn(ctx, "Actualtext with no position. Text may be lost or mispositioned.");
 		pop_metatext(ctx, tdev);
 		return;
 	}
 
+	/* We have a rectangle, so send the text to fill that. */
+	tdev->last.trm.a = tdev->metatext->bounds.x1 - tdev->metatext->bounds.x0;
+	tdev->last.trm.b = 0;
+	tdev->last.trm.c = 0;
+	tdev->last.trm.d = tdev->metatext->bounds.y0 - tdev->metatext->bounds.y1;
+	tdev->last.trm.e = tdev->metatext->bounds.x0;
+	tdev->last.trm.f = tdev->metatext->bounds.y1;
+	tdev->last.valid = 1;
+
 	fz_var(myfont);
 
 	fz_try(ctx)
@@ -1445,7 +1602,7 @@ fz_stext_end_metatext(fz_context *ctx, fz_device *dev)
 			myfont = fz_new_base14_font(ctx, "Helvetica");
 			tdev->last.font = myfont;
 		}
-		flush_actualtext(ctx, tdev, tdev->metatext->text, 0, -1);
+		flush_actualtext(ctx, tdev, tdev->metatext->text, 0, -1, 1);
 		pop_metatext(ctx, tdev);
 	}
 	fz_always(ctx)
@@ -1482,8 +1639,17 @@ fz_stext_fill_image(fz_context *ctx, fz_device *dev, fz_image *img, fz_matrix ct
 
 	/* If the alpha is less than 50% then it's probably a watermark or effect or something. Skip it. */
 	if (alpha >= 0.5f)
-		add_image_block_to_page(ctx, tdev->page, ctm, img, tdev->id);
-
+	{
+		fz_stext_block *block;
+		flush_lazy_vectors(ctx, tdev->page, tdev);
+		block = add_image_block_to_page(ctx, tdev->page, ctm, img, tdev->id);
+		if (tdev->opts.flags & FZ_STEXT_CLIP)
+		{
+			fz_rect clip = fz_device_current_scissor(ctx, dev);
+			clip = fz_intersect_rect(clip, tdev->page->mediabox);
+			block->bbox = fz_intersect_rect(block->bbox, clip);
+		}
+	}
 }
 
 static void
@@ -1534,23 +1700,20 @@ static void
 fz_stext_fill_shade(fz_context *ctx, fz_device *dev, fz_shade *shade, fz_matrix ctm, float alpha, fz_color_params color_params)
 {
 	fz_stext_device *tdev = (fz_stext_device*)dev;
-	fz_rect *bounds = actualtext_bounds(tdev);
 	fz_matrix local_ctm;
 	fz_rect scissor;
 	fz_image *image;
 
-	/* If we aren't keeping images, but we are in a bound, update the bounds
-	 * without generating the entire image. */
-	if ((tdev->opts.flags & FZ_STEXT_PRESERVE_IMAGES) == 0 && bounds)
+	/* If we aren't preserving images, don't waste time making the shade. */
+	if ((tdev->opts.flags & FZ_STEXT_PRESERVE_IMAGES) == 0)
 	{
-		*bounds = fz_union_rect(*bounds, fz_bound_shade(ctx, shade, ctm));
+		/* But we do still need to handle actualtext bounds. */
+		fz_rect *bounds = actualtext_bounds(tdev);
+		if (bounds)
+			*bounds = fz_union_rect(*bounds, fz_bound_shade(ctx, shade, ctm));
 		return;
 	}
 
-	/* Unless we are preserving image, nothing to do here. */
-	if ((tdev->opts.flags & FZ_STEXT_PRESERVE_IMAGES) == 0)
-		return;
-
 	local_ctm = ctm;
 	scissor = fz_device_current_scissor(ctx, dev);
 	if (dev->flags & FZ_STEXT_CLIP_RECT)
@@ -1791,7 +1954,7 @@ check_strikeout(fz_context *ctx, fz_stext_block *block, fz_point from, fz_point
 				/* Dot product with up. up is normalised */
 				dot = dx * up.x + dy * up.y;
 
-				if (dot > 0)
+				if (dot > 0 && dot <= 0.8f * ch->font->ascender * ch->size)
 					ch->flags |= FZ_STEXT_STRIKEOUT;
 				else
 					ch->flags |= FZ_STEXT_UNDERLINE;
@@ -1825,6 +1988,11 @@ fz_stext_close_device(fz_context *ctx, fz_device *dev)
 	fz_stext_device *tdev = (fz_stext_device*)dev;
 	fz_stext_page *page = tdev->page;
 
+	if ((tdev->flags & FZ_STEXT_DEHYPHENATE) && fz_is_unicode_hyphen(tdev->lastchar) && tdev->lastline != NULL)
+		tdev->lastline->flags |= FZ_STEXT_LINE_FLAGS_JOINED;
+
+	flush_lazy_vectors(ctx, page, tdev);
+
 	fixup_bboxes_and_bidi(ctx, page->first_block);
 
 	if (tdev->opts.flags & FZ_STEXT_COLLECT_STYLES)
@@ -1890,6 +2058,8 @@ fz_parse_stext_options(fz_context *ctx, fz_stext_options *opts, const char *stri
 
 	memset(opts, 0, sizeof *opts);
 
+	/* when adding options, remember to update fz_stext_options_usage above */
+
 	if (fz_has_option(ctx, string, "preserve-ligatures", &val) && fz_option_eq(val, "yes"))
 		opts->flags |= FZ_STEXT_PRESERVE_LIGATURES;
 	if (fz_has_option(ctx, string, "preserve-whitespace", &val) && fz_option_eq(val, "yes"))
@@ -1912,6 +2082,8 @@ fz_parse_stext_options(fz_context *ctx, fz_stext_options *opts, const char *stri
 		opts->flags |= FZ_STEXT_ACCURATE_BBOXES;
 	if (fz_has_option(ctx, string, "vectors", &val) && fz_option_eq(val, "yes"))
 		opts->flags |= FZ_STEXT_COLLECT_VECTORS;
+	if (fz_has_option(ctx, string, "lazy-vectors", &val) && fz_option_eq(val, "yes"))
+		opts->flags |= FZ_STEXT_LAZY_VECTORS;
 	if (fz_has_option(ctx, string, "ignore-actualtext", & val) && fz_option_eq(val, "yes"))
 		opts->flags |= FZ_STEXT_IGNORE_ACTUALTEXT;
 	if (fz_has_option(ctx, string, "segment", &val) && fz_option_eq(val, "yes"))
@@ -2117,9 +2289,9 @@ check_for_strikeout(fz_context *ctx, fz_stext_device *tdev, fz_stext_page *page,
 }
 
 static void
-add_vector(fz_context *ctx, fz_stext_page *page, fz_rect bbox, uint32_t flags, uint32_t argb, int id, float exp)
+add_vector(fz_context *ctx, fz_stext_page *page, fz_stext_device *tdev, fz_rect bbox, uint32_t flags, uint32_t argb, int id, float exp)
 {
-	fz_stext_block *b = add_block_to_page(ctx, page, FZ_STEXT_BLOCK_VECTOR, id);
+	fz_stext_block *b;
 
 	if (exp != 0)
 	{
@@ -2129,6 +2301,19 @@ add_vector(fz_context *ctx, fz_stext_page *page, fz_rect bbox, uint32_t flags, u
 		bbox.y1 += exp;
 	}
 
+	if (tdev->flags & (FZ_STEXT_CLIP_RECT | FZ_STEXT_CLIP))
+	{
+		fz_rect r = current_clip(ctx, tdev);
+		bbox = fz_intersect_rect(bbox, r);
+		if (!fz_is_valid_rect(bbox))
+			return;
+	}
+
+	if (tdev->flags & FZ_STEXT_LAZY_VECTORS)
+		b = add_lazy_vector(ctx, page, tdev, id);
+	else
+		b = add_block_to_page(ctx, page, FZ_STEXT_BLOCK_VECTOR, id);
+
 	b->bbox = bbox;
 	b->u.v.flags = flags;
 	b->u.v.argb = argb;
@@ -2136,10 +2321,12 @@ add_vector(fz_context *ctx, fz_stext_page *page, fz_rect bbox, uint32_t flags, u
 
 typedef struct
 {
+	fz_stext_device *dev;
 	fz_matrix ctm;
 	uint32_t argb;
 	uint32_t flags;
 	fz_stext_page *page;
+	fz_rect seg_bounds;
 	fz_rect leftovers;
 	fz_rect pending;
 	int count;
@@ -2153,6 +2340,7 @@ maybe_rect(fz_context *ctx, split_path_data *sp)
 {
 	int rect = 0;
 	int i;
+	fz_rect leftovers;
 
 	if (sp->count >= 0)
 	{
@@ -2183,14 +2371,23 @@ maybe_rect(fz_context *ctx, split_path_data *sp)
 			for (i = 1; i < sp->count; i++)
 				bounds = fz_include_point_in_rect(bounds, sp->p[i]);
 			if (fz_is_valid_rect(sp->pending))
-				add_vector(ctx, sp->page, sp->pending, sp->flags | FZ_STEXT_VECTOR_IS_RECTANGLE | FZ_STEXT_VECTOR_CONTINUES, sp->argb, sp->id, sp->exp);
+				add_vector(ctx, sp->page, sp->dev, sp->pending, sp->flags | FZ_STEXT_VECTOR_IS_RECTANGLE | FZ_STEXT_VECTOR_CONTINUES, sp->argb, sp->id, sp->exp);
 			sp->pending = bounds;
 			return;
 		}
-
-		for (i = 0; i < sp->count; i++)
-			sp->leftovers = fz_include_point_in_rect(sp->leftovers, sp->p[i]);
 	}
+
+	/* We aren't a rectangle! */
+	leftovers = sp->seg_bounds;
+
+	if (sp->dev->flags & (FZ_STEXT_CLIP_RECT | FZ_STEXT_CLIP))
+		leftovers = fz_intersect_rect(leftovers, current_clip(ctx, sp->dev));
+
+	if (fz_is_valid_rect(leftovers))
+		sp->leftovers = fz_union_rect(sp->leftovers, leftovers);
+
+	/* Remember we're not a rect. */
+	sp->count = -1;
 }
 
 static void
@@ -2202,6 +2399,8 @@ split_move(fz_context *ctx, void *arg, float x, float y)
 	maybe_rect(ctx, sp);
 	sp->p[0] = p;
 	sp->count = 1;
+	sp->seg_bounds.x0 = sp->seg_bounds.x1 = p.x;
+	sp->seg_bounds.y0 = sp->seg_bounds.y1 = p.y;
 }
 
 static void
@@ -2209,7 +2408,8 @@ split_line(fz_context *ctx, void *arg, float x, float y)
 {
 	split_path_data *sp = (split_path_data *)arg;
 	fz_point p = fz_transform_point_xy(x, y, sp->ctm);
-	int i;
+
+	sp->seg_bounds = fz_include_point_in_rect(sp->seg_bounds, p);
 
 	if (sp->count >= 0)
 	{
@@ -2234,37 +2434,22 @@ split_line(fz_context *ctx, void *arg, float x, float y)
 				return;
 			}
 		}
-		/* We can no longer be a rect. Output the points we had saved. */
-		for (i = 0; i < sp->count; i++)
-			sp->leftovers = fz_include_point_in_rect(sp->leftovers, sp->p[i]);
-		/* Remember we're not a rect. */
+		/* We can no longer be a rect. */
 		sp->count = -1;
 	}
-	/* Roll this point into the non-rect bounds. */
-	sp->leftovers = fz_include_point_in_rect(sp->leftovers, p);
 }
 
 static void
 split_curve(fz_context *ctx, void *arg, float x1, float y1, float x2, float y2, float x3, float y3)
 {
 	split_path_data *sp = (split_path_data *)arg;
-	fz_point p1 = fz_transform_point_xy(x1, y1, sp->ctm);
-	fz_point p2 = fz_transform_point_xy(x2, y2, sp->ctm);
-	fz_point p3 = fz_transform_point_xy(x3, y3, sp->ctm);
-	int i;
 
-	if (sp->count >= 0)
-	{
-		/* We can no longer be a rect. Output the points we had saved. */
-		for (i = 0; i < sp->count; i++)
-			sp->leftovers = fz_include_point_in_rect(sp->leftovers, sp->p[i]);
-		/* Remember we're not a rect. */
-		sp->count = -1;
-	}
-	/* Roll these points into the non-rect bounds. */
-	sp->leftovers = fz_include_point_in_rect(sp->leftovers, p1);
-	sp->leftovers = fz_include_point_in_rect(sp->leftovers, p2);
-	sp->leftovers = fz_include_point_in_rect(sp->leftovers, p3);
+	sp->seg_bounds = fz_include_point_in_rect(sp->seg_bounds, fz_transform_point_xy(x1, y1, sp->ctm));
+	sp->seg_bounds = fz_include_point_in_rect(sp->seg_bounds, fz_transform_point_xy(x2, y2, sp->ctm));
+	sp->seg_bounds = fz_include_point_in_rect(sp->seg_bounds, fz_transform_point_xy(x3, y3, sp->ctm));
+
+	/* We can no longer be a rect. */
+	sp->count = -1;
 }
 
 static void
@@ -2287,17 +2472,20 @@ fz_path_walker split_path_rects =
 };
 
 static void
-add_vectors_from_path(fz_context *ctx, fz_stext_page *page, int id, const fz_path *path, fz_matrix ctm, fz_colorspace *cs, const float *color, float alpha, fz_color_params cp, int stroke, float exp)
+add_vectors_from_path(fz_context *ctx, fz_stext_page *page, fz_stext_device *tdev, const fz_path *path, fz_matrix ctm, fz_colorspace *cs, const float *color, float alpha, fz_color_params cp, int stroke, float exp)
 {
 	int have_leftovers;
 	split_path_data sp;
+	int id = tdev->id;
 
+	sp.dev = tdev;
 	sp.ctm = ctm;
 	sp.argb = hexrgba_from_color(ctx, cs, color, alpha);
 	sp.flags = stroke ? FZ_STEXT_VECTOR_IS_STROKED : 0;
 	sp.page = page;
 	sp.count = 0;
 	sp.leftovers = fz_empty_rect;
+	sp.seg_bounds = fz_empty_rect;
 	sp.pending = fz_empty_rect;
 	sp.id = id;
 	sp.exp = exp;
@@ -2308,9 +2496,9 @@ add_vectors_from_path(fz_context *ctx, fz_stext_page *page, int id, const fz_pat
 	maybe_rect(ctx, &sp);
 
 	if (fz_is_valid_rect(sp.pending))
-		add_vector(ctx, page, sp.pending, sp.flags | FZ_STEXT_VECTOR_IS_RECTANGLE | (have_leftovers ? FZ_STEXT_VECTOR_CONTINUES : 0), sp.argb, id, exp);
+		add_vector(ctx, page, sp.dev, sp.pending, sp.flags | FZ_STEXT_VECTOR_IS_RECTANGLE | (have_leftovers ? FZ_STEXT_VECTOR_CONTINUES : 0), sp.argb, id, exp);
 	if (have_leftovers)
-		add_vector(ctx, page, sp.leftovers, sp.flags, sp.argb, id, exp);
+		add_vector(ctx, page, sp.dev, sp.leftovers, sp.flags, sp.argb, id, exp);
 }
 
 static void
@@ -2329,7 +2517,7 @@ fz_stext_fill_path(fz_context *ctx, fz_device *dev, const fz_path *path, int eve
 		check_for_strikeout(ctx, tdev, page, path, ctm);
 
 	if (tdev->flags & FZ_STEXT_COLLECT_VECTORS)
-		add_vectors_from_path(ctx, page, tdev->id, path, ctm, cs, color, alpha, cp, 0, 0);
+		add_vectors_from_path(ctx, page, tdev, path, ctm, cs, color, alpha, cp, 0, 0);
 }
 
 static void
@@ -2349,7 +2537,7 @@ fz_stext_stroke_path(fz_context *ctx, fz_device *dev, const fz_path *path, const
 		check_for_strikeout(ctx, tdev, page, path, ctm);
 
 	if (tdev->flags & FZ_STEXT_COLLECT_VECTORS)
-		add_vectors_from_path(ctx, page, tdev->id, path, ctm, cs, color, alpha, cp, 1, exp);
+		add_vectors_from_path(ctx, page, tdev, path, ctm, cs, color, alpha, cp, 1, exp);
 }
 
 static void
@@ -2487,6 +2675,19 @@ fz_stext_begin_structure(fz_context *ctx, fz_device *dev, fz_structure standard,
 		newblock->prev = gt->prev;
 		if (gt->prev)
 			gt->prev->next = newblock;
+		else if (page->last_struct)
+		{
+			/* We're linking it in at the start under another struct! */
+			assert(page->last_struct->first_block == gt);
+			assert(page->last_struct->last_block != NULL);
+			page->last_struct->first_block = newblock;
+		}
+		else
+		{
+			/* We're linking it in at the start of the page! */
+			assert(page->first_block == gt);
+			page->first_block = newblock;
+		}
 		gt->prev = newblock;
 		newblock->next = gt;
 		newblock->id = gt->id;
@@ -2598,13 +2799,16 @@ fz_new_stext_device_for_page(fz_context *ctx, fz_stext_page *page, const fz_stex
 	dev->pen.y = 0;
 	dev->trm = fz_identity;
 	dev->lastchar = ' ';
+	dev->lastline = NULL;
 	dev->lasttext = NULL;
 	dev->lastbidi = 0;
 	dev->last_was_fake_bold = 1;
 	if (opts)
 		dev->opts = *opts;
 
-	if ((dev->flags & FZ_STEXT_PRESERVE_IMAGES) == 0)
+	/* If we are ignoring images, then it'd be nice to skip the decode costs. BUT we still need them to tell
+	 * us the bounds for ActualText, so we can only actually skip them if we are ignoring actualtext too. */
+	if ((dev->flags & FZ_STEXT_PRESERVE_IMAGES) == 0 && (dev->opts.flags & FZ_STEXT_IGNORE_ACTUALTEXT) != 0)
 		dev->super.hints |= FZ_DONT_DECODE_IMAGES;
 
 	dev->rect_max = 0;
