@@ -1012,7 +1012,7 @@ walk_to_find_content(fz_context *ctx, div_list *xs, div_list *ys, fz_stext_block
 				if (region.y1 > bounds.y1)
 					region.y1 = bounds.y1;
 				if (region.y0 >= region.y1)
-					break;
+					continue;
 
 				/* Skip leading spaces. */
 				for (ch = line->first_char; ch != NULL; ch = ch->next)
@@ -1166,6 +1166,37 @@ typedef struct
 	int has_background;
 } grid_walker_data;
 
+static fz_stext_grid_info *
+copy_grid_info_to_pool(fz_context *ctx, fz_stext_page *page, cells_t *cells)
+{
+	int i, n = cells->w * cells->h;
+	fz_stext_grid_info *info;
+	size_t z = offsetof(fz_stext_grid_info, info) + sizeof(info->info[0]) * n;
+	cell_t *cell = &cells->cell[0];
+	info = fz_pool_alloc(ctx, page->pool, z);
+
+	info->w = cells->w;
+	info->h = cells->h;
+
+	for (i = 0; i < n; i++, cell++)
+	{
+		unsigned int flags = 0;
+		if (cell->full)
+			flags |= FZ_STEXT_GRID_FULL;
+		if (cell->h_crossed)
+			flags |= FZ_STEXT_GRID_H_CROSSED;
+		if (cell->v_crossed)
+			flags |= FZ_STEXT_GRID_V_CROSSED;
+		if (cell->h_line)
+			flags |= FZ_STEXT_GRID_T_BORDER;
+		if (cell->v_line)
+			flags |= FZ_STEXT_GRID_L_BORDER;
+		info->info[i].flags = flags;
+	}
+
+	return info;
+}
+
 static cell_t *
 get_cell(cells_t *cells, int x, int y)
 {
@@ -1349,10 +1380,24 @@ find_grid_pos(fz_context *ctx, grid_walker_data *gd, int row, float x, int inacc
 		else if (x <= pos->list[i].max)
 		{
 			/* We are in the range for the ith divider. */
-			if (i == 0 || i == pos->len-1)
+			if (i == 0)
 			{
-				/* Never move the outermost pos in, because they have been
-				 * calculated to be just big enough already. */
+				/* Never move the first pos down, because it's been
+				 * calculated to be just big enough already. But we might
+				 * have to split it. */
+				if (pos->list[0].pos + WIGGLE_ROOM < x)
+					goto split;
+
+				return 0;
+			}
+			if (i == pos->len-1)
+			{
+				/* Never move the last pos in, because it's been
+				 * calculated to be just big enough already. But we might
+				 * have to split it. */
+				if (pos->list[i].pos - WIGGLE_ROOM > x)
+					goto split;
+
 				return i;
 			}
 			if (pos->list[i].reinforcement == 0)
@@ -1373,6 +1418,7 @@ find_grid_pos(fz_context *ctx, grid_walker_data *gd, int row, float x, int inacc
 				return i;
 			}
 			/* We need to split i into i and i+1. */
+split:
 			pos = split_grid_pos(ctx, gd, row, i, pos->list[i].pos > x);
 			if (pos->list[i].pos > x)
 			{
@@ -1427,43 +1473,71 @@ find_cell_r(fz_stext_grid_positions *pos, float x)
 	return -1;
 }
 
-/* Add a horizontal line. Return 1 if the line doesn't seem to be a border line.
- * Record which cells that was a border for. */
+/* Lookup the position of a horizontal line, reinforcing as appropriate. */
 static void
 add_h_line(fz_context *ctx, grid_walker_data *gd, float x0, float x1, float y0, float y1)
+{
+	float y = (y0 + y1) / 2;
+	(void)find_grid_pos(ctx, gd, 1, y, 0);
+}
+
+/* Lookup the position of a vertical line, reinforcing as appropriate. */
+static void
+add_v_line(fz_context *ctx, grid_walker_data *gd, float y0, float y1, float x0, float x1)
+{
+	float x = (x0 + x1) / 2;
+	(void)find_grid_pos(ctx, gd, 0, x, 0);
+}
+
+/* Lookup the position of a rectangle, reinforcing as appropriate. */
+static void
+add_hv_line(fz_context *ctx, grid_walker_data *gd, float x0, float x1, float y0, float y1, int stroked)
+{
+	(void)find_grid_pos(ctx, gd, 0, x0, 0);
+	(void)find_grid_pos(ctx, gd, 0, x1, 0);
+	(void)find_grid_pos(ctx, gd, 1, y0, 0);
+	(void)find_grid_pos(ctx, gd, 1, y1, 0);
+}
+
+/* Add a horizontal line (with no reinforcement).
+ * Record which cells that was a border for. */
+static void
+add_h_line2(fz_context *ctx, grid_walker_data *gd, float x0, float x1, float y0, float y1)
 {
 	int start = find_grid_pos(ctx, gd, 0, x0, 1);
 	int end = find_grid_pos(ctx, gd, 0, x1, 1);
 	float y = (y0 + y1) / 2;
-	int yidx = find_grid_pos(ctx, gd, 1, y, 0);
+	int yidx = find_grid_pos(ctx, gd, 1, y, 1);
 	int i;
 
 	for (i = start; i < end; i++)
 		get_cell(gd->cells, i, yidx)->h_line++;
 }
 
-/* Add a vertical line. Return 1 if the line doesn't seem to be a border line.
+/* Add a vertical line (with no reinforcement).
  * Record which cells that was a border for. */
 static void
-add_v_line(fz_context *ctx, grid_walker_data *gd, float y0, float y1, float x0, float x1)
+add_v_line2(fz_context *ctx, grid_walker_data *gd, float y0, float y1, float x0, float x1)
 {
 	int start = find_grid_pos(ctx, gd, 1, y0, 1);
 	int end = find_grid_pos(ctx, gd, 1, y1, 1);
 	float x = (x0 + x1) / 2;
-	int xidx = find_grid_pos(ctx, gd, 0, x, 0);
+	int xidx = find_grid_pos(ctx, gd, 0, x, 1);
 	int i;
 
 	for (i = start; i < end; i++)
 		get_cell(gd->cells, xidx, i)->v_line++;
 }
 
+/* Add a rectangle (with no reinforcement).
+ * Record which cells that was a border for. */
 static void
-add_hv_line(fz_context *ctx, grid_walker_data *gd, float x0, float x1, float y0, float y1, int stroked)
+add_hv_line2(fz_context *ctx, grid_walker_data *gd, float x0, float x1, float y0, float y1, int stroked)
 {
-	int ix0 = find_grid_pos(ctx, gd, 0, x0, 0);
-	int ix1 = find_grid_pos(ctx, gd, 0, x1, 0);
-	int iy0 = find_grid_pos(ctx, gd, 1, y0, 0);
-	int iy1 = find_grid_pos(ctx, gd, 1, y1, 0);
+	int ix0 = find_grid_pos(ctx, gd, 0, x0, 1);
+	int ix1 = find_grid_pos(ctx, gd, 0, x1, 1);
+	int iy0 = find_grid_pos(ctx, gd, 1, y0, 1);
+	int iy1 = find_grid_pos(ctx, gd, 1, y1, 1);
 	int i;
 
 	if (stroked)
@@ -1667,7 +1741,7 @@ walk_grid_lines(fz_context *ctx, grid_walker_data *gd, fz_stext_block *block)
 			h = r.y1 - r.y0;
 			if (w > h && h < 2)
 			{
-				/* Thin, wide line */
+				/* Thin, tall line */
 				(void) add_h_line(ctx, gd, r.x0, r.x1, r.y0, r.y1);
 			}
 			else if (w < h && w < 2)
@@ -1684,6 +1758,52 @@ walk_grid_lines(fz_context *ctx, grid_walker_data *gd, fz_stext_block *block)
 	}
 }
 
+static void
+walk_grid_lines2(fz_context *ctx, grid_walker_data *gd, fz_stext_block *block)
+{
+	for (; block != NULL; block = block->next)
+	{
+		if (block->type == FZ_STEXT_BLOCK_STRUCT)
+		{
+			if (block->u.s.down)
+				walk_grid_lines2(ctx, gd, block->u.s.down->first_block);
+			continue;
+		}
+		else if (block->type == FZ_STEXT_BLOCK_VECTOR)
+		{
+			fz_rect r;
+			float w, h;
+
+			/* Only process rectangle blocks. */
+			if ((block->u.v.flags & FZ_STEXT_VECTOR_IS_RECTANGLE) == 0)
+				continue;
+
+			r = fz_collate_small_vector_run(&block);
+			r = fz_intersect_rect(r, gd->bounds);
+			if (!fz_is_valid_rect(r))
+				continue;
+
+			w = r.x1 - r.x0;
+			h = r.y1 - r.y0;
+			if (w > h && h < 2)
+			{
+				/* Thin, wide line */
+				(void) add_h_line2(ctx, gd, r.x0, r.x1, r.y0, r.y1);
+			}
+			else if (w < h && w < 2)
+			{
+				/* Thin, wide line */
+				(void) add_v_line2(ctx, gd, r.y0, r.y1, r.x0, r.x1);
+			}
+			else
+			{
+				/* Rectangle */
+				(void) add_hv_line2(ctx, gd, r.x0, r.x1, r.y0, r.y1, block->u.v.flags & FZ_STEXT_VECTOR_IS_STROKED);
+			}
+		}
+	}
+}
+
 static int
 is_numeric(int c)
 {
@@ -1696,7 +1816,10 @@ mark_cells_for_content(fz_context *ctx, grid_walker_data *gd, fz_rect s)
 	fz_rect r = fz_intersect_rect(gd->bounds, s);
 	int x0, x1, y0, y1, x, y;
 
-	if (fz_is_empty_rect(r))
+	/* Check for non-validity rather than empty here, as e.g.
+	 * spaces are empty, and we'd still like to account for
+	 * their horizontal extent. */
+	if (!fz_is_valid_rect(r))
 		return 0;
 
 	x0 = find_cell_l(gd->xpos, r.x0);
@@ -1934,6 +2057,11 @@ calculate_spanned_content(fz_context *ctx, grid_walker_data *gd, fz_stext_block
 							was_numeric = 0;
 							continue;
 						}
+						if (ch->flags & FZ_STEXT_SYNTHETIC_LARGE)
+						{
+							/* Break on large synthetic spaces */
+							continue;
+						}
 						/* A single space. Accept it. */
 						was_numeric = 0;
 					}
@@ -2364,16 +2492,14 @@ find_table_insertion_point(fz_context *ctx, fz_rect r, tree_pos current, tree_po
 				down.parent = block->u.s.down;
 				down.block = block->u.s.down->first_block;
 				best = find_table_insertion_point(ctx, r, down, best);
+				continue;
 			}
 		}
-		else
+		/* Is block a better precursor than best? (Or a valid precursor, if best.block == NULL) */
+		if (block->bbox.y1 < r.y0 && (best.block == NULL || best.block->bbox.y1 < block->bbox.y1))
 		{
-			/* Is block a better precursor than best? (Or a valid precursor, if best.block == NULL) */
-			if (block->bbox.y1 < r.y0 && (best.block == NULL || best.block->bbox.y1 < block->bbox.y1))
-			{
-				best.block = block;
-				best.parent = current.parent;
-			}
+			best.block = block;
+			best.parent = current.parent;
 		}
 	}
 
@@ -2595,7 +2721,8 @@ transcribe_table(fz_context *ctx, grid_walker_data *gd, fz_stext_page *page, fz_
 	r.y0 = gd->ypos->list[0].pos;
 	r.y1 = gd->ypos->list[h-1].pos;
 #ifdef DEBUG_TABLE_STRUCTURE
-	fz_print_stext_page_as_xml(ctx, fz_stddbg(ctx), page, 0);
+	printf("Region %g %g %g %g\n", r.x0, r.y0, r.x1, r.y1);
+	fz_print_stext_page_as_xml(ctx, fz_stdout(ctx), page, 0);
 #endif
 	top.block = *first_block;
 	top.parent = parent;
@@ -2714,9 +2841,11 @@ transcribe_table(fz_context *ctx, grid_walker_data *gd, fz_stext_page *page, fz_
 		fz_stext_block *block;
 		fz_stext_grid_positions *xps2 = copy_grid_positions_to_pool(ctx, page, gd->xpos);
 		fz_stext_grid_positions *yps2 = copy_grid_positions_to_pool(ctx, page, gd->ypos);
+		fz_stext_grid_info *info = copy_grid_info_to_pool(ctx, page, gd->cells);
 		block = add_grid_block(ctx, page, &table->first_block, &table->last_block, table->up->id);
 		block->u.b.xs = xps2;
 		block->u.b.ys = yps2;
+		block->u.b.info = info;
 		block->bbox.x0 = block->u.b.xs->list[0].pos;
 		block->bbox.y0 = block->u.b.ys->list[0].pos;
 		block->bbox.x1 = block->u.b.xs->list[block->u.b.xs->len-1].pos;
@@ -2737,19 +2866,19 @@ merge_column(grid_walker_data *gd, int x)
 		cell_t *s = &gd->cells->cell[x + y * gd->cells->w];
 
 		if (x > 0)
-			memcpy(d-x, s-x, sizeof(*d) * x);
+			memmove(d-x, s-x, sizeof(*d) * x);
 		d->full = s[0].full || s[1].full;
 		d->h_crossed = s[0].h_crossed || s[1].h_crossed;
 		d->h_line = s[0].h_line; /* == s[1].h_line */
 		d->v_crossed = s[0].v_crossed;
 		d->v_line = s[0].v_line;
 		if (x < gd->cells->w - 2)
-			memcpy(d+1, s+2, sizeof(*d) * (gd->cells->w - 2 - x));
+			memmove(d+1, s+2, sizeof(*d) * (gd->cells->w - 2 - x));
 	}
 	gd->cells->w--;
 
 	if (x < gd->xpos->len - 2)
-		memcpy(&gd->xpos->list[x+1], &gd->xpos->list[x+2], sizeof(gd->xpos->list[0]) * (gd->xpos->len - 2 - x));
+		memmove(&gd->xpos->list[x+1], &gd->xpos->list[x+2], sizeof(gd->xpos->list[0]) * (gd->xpos->len - 2 - x));
 	gd->xpos->len--;
 }
 
@@ -2840,11 +2969,11 @@ merge_row(grid_walker_data *gd, int y)
 		d++;
 	}
 	if (y < gd->cells->h - 2)
-		memcpy(d, d+w, sizeof(*d) * (gd->cells->h - 2 - y) * w);
+		memmove(d, d+w, sizeof(*d) * (gd->cells->h - 2 - y) * w);
 	gd->cells->h--;
 
 	if (y < gd->ypos->len - 2)
-		memcpy(&gd->ypos->list[y+1], &gd->ypos->list[y+2], sizeof(gd->ypos->list[0]) * (gd->ypos->len - 2 - y));
+		memmove(&gd->ypos->list[y+1], &gd->ypos->list[y+2], sizeof(gd->ypos->list[0]) * (gd->ypos->len - 2 - y));
 	gd->ypos->len--;
 }
 
@@ -3110,8 +3239,11 @@ find_table_within_bounds(fz_context *ctx, grid_walker_data *gd, fz_stext_block *
 			gd->has_background = 1;
 
 		/* Walk the content looking for grid lines. These
-		 * lines refine our positions. */
+		 * lines refine our positions. 2 passes. First pass
+		 * ensures that all the lines are in the grid. Second
+		 * pass then marks them. */
 		walk_grid_lines(ctx, gd, content);
+		walk_grid_lines2(ctx, gd, content);
 		/* Now, we walk the content looking for content that crosses
 		 * these grid lines. This allows us to spot spanned cells. */
 		if (calculate_spanned_content(ctx, gd, content))
@@ -3147,14 +3279,145 @@ find_table_within_bounds(fz_context *ctx, grid_walker_data *gd, fz_stext_block *
 	return failed;
 }
 
+/* Is cell (x,y) plausibly part of a bordered cell (or 'super-cell'
+ * allowing for spanning)? */
 static int
-do_table_hunt(fz_context *ctx, fz_stext_page *page, fz_stext_struct *parent, int *has_background, fz_rect top_bounds)
+plausibly_bordered_spanned_cell(cells_t *cells, int x, int y)
+{
+	int minx, miny, maxx, maxy;
+
+	for (minx = x; minx >= 0; minx--)
+	{
+		cell_t *cell = get_cell(cells, minx, y);
+		if (cell->v_line)
+			break;
+	}
+	if (minx < 0)
+		return 0;
+	for (maxx = x+1; maxx < cells->w-1; maxx++)
+	{
+		cell_t *cell = get_cell(cells, maxx, y);
+		if (cell->v_line)
+			break;
+	}
+	if (maxx == cells->w)
+		return 0;
+	for (miny = y; miny >= 0; miny--)
+	{
+		cell_t *cell = get_cell(cells, x, miny);
+		if (cell->h_line)
+			break;
+	}
+	if (miny < 0)
+		return 0;
+	for (maxy = y+1; maxy < cells->h-1; maxy++)
+	{
+		cell_t *cell = get_cell(cells, x, maxy);
+		if (cell->h_line)
+			break;
+	}
+	if (maxy == cells->h)
+		return 0;
+
+	/* So we know have a plausible size for this large cell. */
+	/* Now check that borders exist everywhere we expect. */
+	for (x = minx; x < maxx; x++)
+	{
+		cell_t *cell = get_cell(cells, x, miny);
+		if (!cell->h_line)
+			return 0;
+		cell = get_cell(cells, x, maxy);
+		if (!cell->h_line)
+			return 0;
+	}
+	for (y = miny; y < maxy; y++)
+	{
+		cell_t *cell = get_cell(cells, minx, y);
+		if (!cell->v_line)
+			return 0;
+		cell = get_cell(cells, maxx, y);
+		if (!cell->v_line)
+			return 0;
+	}
+
+	/* And now nowhere we don't. */
+	for (y = miny+1; y < maxy-1; y++)
+		for (x = minx+1; x < maxx-1; x++)
+		{
+			cell_t *cell = get_cell(cells, x, y);
+			if (cell->h_line || cell->v_line)
+				return 0;
+		}
+
+	return 1;
+}
+
+/* The score for a table can be thought of as a judgement of
+ * how 'awkward' a table is.
+ *
+ *  + Score 1 for every empty cell that isn't part of a properly
+ *    bordered region, (or 0.5 if it's 'half' bordered)
+ *  + Score 1 for every 'crossing' between cells.
+ */
+static float
+score_table(fz_context *ctx, grid_walker_data *gd)
+{
+	int x, y;
+	int w = gd->cells->w;
+	int h = gd->cells->h;
+	int score = 0;
+	int num_cells = (w-1)*(h-1);
+
+	assert(num_cells > 0);
+
+	for (y = 0; y < h-1; y++)
+	{
+		for (x = 0; x < w-1; x++)
+		{
+			cell_t *cell = get_cell(gd->cells, x, y);
+			cell_t *right = get_cell(gd->cells, x+1, y);
+			cell_t *below = get_cell(gd->cells, x, y+1);
+			score += cell->h_crossed + cell->v_crossed;
+			if (cell->full)
+			{
+				/* We have content. */
+			}
+			else if (cell->h_line && cell->v_line && right->v_line && below->h_line)
+			{
+				/* The cell is trivially bordered. */
+			}
+			else if (plausibly_bordered_spanned_cell(gd->cells, x, y))
+			{
+			}
+			else if (cell->h_line && below->h_line)
+			{
+				/* The cell is halfway bordered. */
+				score += 1;
+			}
+			else if (cell->v_line && right->v_line)
+			{
+				/* The cell is halfway bordered. */
+				score += 1;
+			}
+			else
+				score += 2;
+		}
+	}
+
+	return score / (float)num_cells / 2.0;
+}
+
+static int
+do_table_hunt(fz_context *ctx, fz_stext_page *page, fz_stext_struct *parent, int *has_background, fz_rect top_bounds, float *subtable_score, float score_threshold)
 {
 	fz_stext_block *block;
 	int count;
 	fz_stext_block **first_block = parent ? &parent->first_block : &page->first_block;
 	int num_subtables = 0;
 	grid_walker_data gd = { 0 };
+	float score;
+
+	*subtable_score = 0;
 
 	/* No content? Just bale. */
 	if (*first_block == NULL)
@@ -3173,12 +3436,22 @@ do_table_hunt(fz_context *ctx, fz_stext_page *page, fz_stext_struct *parent, int
 	{
 		if (block->type == FZ_STEXT_BLOCK_STRUCT)
 		{
-			if (block->u.s.down)
+			if (block->u.s.down && fz_is_valid_rect(fz_intersect_rect(top_bounds, block->bbox)))
 			{
+				float my_subtable_score;
 				int background = 0;
-				num_subtables += do_table_hunt(ctx, page, block->u.s.down, &background, top_bounds);
+				int st;
+#ifdef DEBUG_TABLE_STRUCTURE
+				printf("Looking for subtable\n");
+#endif
+				st = do_table_hunt(ctx, page, block->u.s.down, &background, top_bounds, &my_subtable_score, score_threshold);
+				num_subtables += st;
+#ifdef DEBUG_TABLE_STRUCTURE
+				printf("subtable scored %g in %d subtables\n", my_subtable_score, st);
+#endif
 				if (background)
 					*has_background = 1;
+				*subtable_score += my_subtable_score;
 				count++;
 			}
 		}
@@ -3211,28 +3484,63 @@ do_table_hunt(fz_context *ctx, fz_stext_page *page, fz_stext_struct *parent, int
 		/* Now see whether the content looks like tables. */
 		fz_rect bounds = walk_to_find_bounds(ctx, *first_block);
 
+		bounds = fz_intersect_rect(bounds, top_bounds);
+
 		if (find_table_within_bounds(ctx, &gd, *first_block, bounds))
 			break;
 
+		score = score_table(ctx, &gd);
+#ifdef DEBUG_TABLE_STRUCTURE
+		printf("Bounds: %g %g %g %g Score: %g\n",
+			bounds.x0, bounds.y0, bounds.x1, bounds.y1, score);
+#endif
+
 		*has_background = gd.has_background;
 
 		if (num_subtables > 0)
 		{
+			/* Our table's score needs to be lower than the sum of the scores
+			 * for the subtables for us to accept it. */
+			int x, y, ditch;
+			if (score > *subtable_score)
+			{
+#ifdef DEBUG_TABLE_STRUCTURE
+				printf("Ignoring table as score (%g) > subtable_score (%g)\n", score, *subtable_score);
+#endif
+				break;
+			}
 			/* We are risking throwing away a table we've already found for this
 			 * one. Only do it if this one is really convincing. */
-			int x, y;
+			ditch = 0;
 			for (x = 0; x < gd.xpos->len; x++)
 				if (gd.xpos->list[x].uncertainty != 0)
-					break;
+					ditch = 1;
 			if (x != gd.xpos->len)
-				break;
+				ditch = 2;
 			for (y = 0; y < gd.ypos->len; y++)
 				if (gd.ypos->list[y].uncertainty != 0)
-					break;
+					ditch = 3;
 			if (y != gd.ypos->len)
+				ditch = 4;
+			if (ditch)
+			{
+#ifdef DEBUG_TABLE_STRUCTURE
+				printf("Ignoring table as it seems uncertain (%d).\n", ditch);
+#endif
 				break;
+			}
 		}
 
+		if (score > score_threshold)
+		{
+#ifdef DEBUG_TABLE_STRUCTURE
+			printf("Ignoring table as score (%g) > score_threshold (%g)\n", score, score_threshold);
+#endif
+			break;
+		}
+
+		*subtable_score = score;
+
 #ifdef DEBUG_TABLE_STRUCTURE
 		printf("Transcribing table: (%g,%g)->(%g,%g)\n",
 			gd.xpos->list[0].pos,
@@ -3241,7 +3549,6 @@ do_table_hunt(fz_context *ctx, fz_stext_page *page, fz_stext_struct *parent, int
 			gd.ypos->list[gd.ypos->len-1].pos);
 #endif
 
-		/* Now we should have the entire table calculated. */
 		(void)transcribe_table(ctx, &gd, page, parent);
 		num_subtables = 1;
 #ifdef DEBUG_WRITE_AS_PS
@@ -3287,13 +3594,40 @@ void
 fz_table_hunt(fz_context *ctx, fz_stext_page *page)
 {
 	int has_background = 0;
+	float score;
+	fz_flotilla *flot;
+	int found = 0;
 
 	if (page == NULL)
 		return;
 
 	assert(verify_stext(ctx, page, NULL));
 
-	do_table_hunt(ctx, page, NULL, &has_background, fz_infinite_rect);
+	flot = fz_new_flotilla_from_stext_page_vectors(ctx, page);
+	fz_try(ctx)
+	{
+		int i;
+		int n = fz_flotilla_size(ctx, flot);
+
+		for (i = 0; i < n; i++)
+		{
+			fz_rect r = fz_flotilla_raft_area(ctx, flot, i);
+
+			/* FIXME: Nasty heuristic threshold. */
+			found += do_table_hunt(ctx, page, NULL, &has_background, r, &score, 0.8f);
+		}
+	}
+	fz_always(ctx)
+		fz_drop_flotilla(ctx, flot);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+
+	/* If we can't find any that way, fall back to hunting within segments. */
+	if (found == 0)
+	{
+		/* FIXME: Nasty heuristic threshold. */
+		do_table_hunt(ctx, page, NULL, &has_background, fz_infinite_rect, &score, 0.3f);
+	}
 
 	assert(verify_stext(ctx, page, NULL));
 }
@@ -3302,6 +3636,7 @@ void
 fz_table_hunt_within_bounds(fz_context *ctx, fz_stext_page *page, fz_rect rect)
 {
 	int has_background = 0;
+	float score;
 
 	if (page == NULL)
 		return;
@@ -3310,7 +3645,7 @@ fz_table_hunt_within_bounds(fz_context *ctx, fz_stext_page *page, fz_rect rect)
 
 	fz_segment_stext_rect(ctx, page, rect);
 
-	do_table_hunt(ctx, page, NULL, &has_background, rect);
+	do_table_hunt(ctx, page, NULL, &has_background, rect, &score, 1);
 
 	assert(verify_stext(ctx, page, NULL));
 }
@@ -3344,8 +3679,10 @@ fz_find_table_within_bounds(fz_context *ctx, fz_stext_page *page, fz_rect bounds
 		/* Now we should have the entire table calculated. */
 		table = transcribe_table(ctx, &gd, page, NULL);
 #ifdef DEBUG_WRITE_AS_PS
+		if (table && table->first_block && table->first_block->type == FZ_STEXT_BLOCK_GRID)
 		{
 			int i;
+			fz_stext_block *block = table->first_block;
 			printf("%% TABLE\n");
 			for (i = 0; i < block->u.b.xs->len; i++)
 			{
